use std::{
    fmt::{Debug, Display, Formatter},
    path::PathBuf,
};

#[derive(Debug)]
pub enum Error {
    IOError(Box<dyn std::error::Error>),
    InvalidObjectType(u32),
    InvalidSectionName,
    InvalidSectionData,
    ParseError(Box<dyn std::error::Error>),
    TryFromIntError,
    LinkingError(Trace),
}

// "backtrace" for error origin
#[derive(Debug, Default)]
pub struct Trace {
    pub code: u32, // specific trace code, should be >= 700
    pub message: &'static str,
    pub origin: PathBuf,
    pub offset: u64, // 0 indicates unknown/invalid
    pub source_info: Option<Box<dyn SourceInfo>>,
}

impl Error {
    // used for programmatic input/output handling, IDE's etc.
    pub fn code(&self) -> u32 {
        match self {
            Error::IOError(_) => 100,
            Error::InvalidObjectType(_) => 200,
            Error::InvalidSectionName => 300,
            Error::InvalidSectionData => 400,
            Error::ParseError(_) => 500,
            Error::TryFromIntError => 600,
            Error::LinkingError(trace) => trace.code,
        }
    }
}

pub trait SourceInfo: Debug + Display {
    fn line(&self) -> u64;

    // todo
}

impl From<std::io::Error> for Error {
    fn from(err: std::io::Error) -> Self {
        Self::IOError(Box::new(err))
    }
}

impl From<std::num::TryFromIntError> for Error {
    fn from(_: std::num::TryFromIntError) -> Self {
        Self::TryFromIntError
    }
}

impl Display for Error {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        write!(f, "[{}] ", self.code())?;

        match self {
            Error::IOError(err) => write!(f, "IO error: {}", err),
            Error::InvalidObjectType(ot) => write!(f, "Invalid object type: {}", ot),
            Error::InvalidSectionName => write!(f, "Invalid section name"),
            Error::InvalidSectionData => write!(f, "Invalid section data"),
            Error::ParseError(err) => write!(f, "Parse error: {}", err),
            Error::TryFromIntError => write!(f, "Integer conversion error"),
            Error::LinkingError(trace) => write!(f, "Linking error: {}", trace),
        }
    }
}

impl Display for Trace {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        let mut new_line = ""; // new line in case of previous info only

        if let Some(origin) = self.origin.to_str() {
            new_line = "\n";
            write!(f, "\n\t{}", origin)?;
        }

        if self.offset > 0 {
            new_line = "\n";
            write!(f, "\n@offset: {}", self.offset)?;
        }

        if let Some(si) = &self.source_info {
            new_line = "\n";
            write!(f, "\n{}", si)?;
        }

        write!(f, "{}{}", new_line, self.message)
    }
}
